from libqtile.utils import lazify_imports
from libqtile.widget.import_error import make_error

widgets = {
    "AGroupBox": "groupbox",
    "Backlight": "backlight",
    "Battery": "battery",
    "BatteryIcon": "battery",
    "Bluetooth": "bluetooth",
    "CPU": "cpu",
    "CPUGraph": "graph",
    "Canto": "canto",
    "CapsNumLockIndicator": "caps_num_lock_indicator",
    "CheckUpdates": "check_updates",
    "Chord": "chord",
    "Clipboard": "clipboard",
    "Clock": "clock",
    "Countdown": "countdown",
    "CryptoTicker": "crypto_ticker",
    "CurrentLayout": "currentlayout",
    "CurrentScreen": "currentscreen",
    "DF": "df",
    "DoNotDisturb": "do_not_disturb",
    "GenPollText": "generic_poll_text",
    "GenPollUrl": "gen_poll_url",
    "GenPollCommand": "generic_poll_text",
    "GmailChecker": "gmail_checker",
    "GroupBox": "groupbox",
    "HDD": "hdd",
    "HDDBusyGraph": "graph",
    "HDDGraph": "graph",
    "IdleRPG": "idlerpg",
    "Image": "image",
    "ImapWidget": "imapwidget",
    "KeyboardKbdd": "keyboardkbdd",
    "KeyboardLayout": "keyboardlayout",
    "KhalCalendar": "khal_calendar",
    "LaunchBar": "launchbar",
    "Load": "load",
    "Maildir": "maildir",
    "Memory": "memory",
    "MemoryGraph": "graph",
    "Mirror": "base",
    "Mpd2": "mpd2widget",
    "Mpris2": "mpris2widget",
    "Net": "net",
    "NetGraph": "graph",
    "NetUP": "netup",
    "Notify": "notify",
    "NvidiaSensors": "nvidia_sensors",
    "OpenWeather": "open_weather",
    "Plasma": "plasma",
    "Pomodoro": "pomodoro",
    "Prompt": "prompt",
    "PulseVolume": "pulse_volume",
    "QuickExit": "quick_exit",
    "Redshift": "redshift",
    "ScreenSplit": "screensplit",
    "Sep": "sep",
    "She": "she",
    "Spacer": "spacer",
    "StatusNotifier": "statusnotifier",
    "StockTicker": "stock_ticker",
    "SwapGraph": "graph",
    "SwayNC": "swaync",
    "Systray": "systray",
    "TaskList": "tasklist",
    "TextBox": "textbox",
    "ThermalSensor": "sensors",
    "ThermalZone": "thermal_zone",
    "TunedManager": "tuned_manager",
    "VerticalClock": "vertical_clock",
    "Volume": "volume",
    "Wallpaper": "wallpaper",
    "WidgetBox": "widgetbox",
    "WindowCount": "window_count",
    "WindowName": "windowname",
    "WindowTabs": "windowtabs",
    "Wlan": "wlan",
    "WlanIw": "wlaniw",
    "Wttr": "wttr",
}

__all__, __dir__, __getattr__ = lazify_imports(widgets, __package__, fallback=make_error)
